@@ -2,39 +2,24 @@
 """
 import zmq
 from zmq import *
+from zmq import devices
+__all__ = zmq.__all__
 
-# imported with different names as to not have the star import try to to clobber (when building with cython)
-from zmq.core.context import Context as _original_Context
-from zmq.core.socket import Socket as _original_Socket
-
+import gevent
+from gevent import select
 from gevent.event import AsyncResult
 from gevent.hub import get_hub
 
 
-class _Context(_original_Context):
-    """Replacement for :class:`zmq.core.context.Context`
-
-    Ensures that the greened Socket below is used in calls to `socket`.
-    """
-
-    def socket(self, socket_type):
-        """Overridden method to ensure that the green version of socket is used
-
-        Behaves the same as :meth:`zmq.core.context.Context.socket`, but ensures
-        that a :class:`Socket` with all of its send and recv methods set to be
-        non-blocking is returned
-        """
-        if self.closed:
-            raise ZMQError(ENOTSUP)
-        return _Socket(self, socket_type)
-
-class _Socket(_original_Socket):
+class GreenSocket(Socket):
     """Green version of :class:`zmq.core.socket.Socket`
 
     The following methods are overridden:
 
         * send
+        * send_multipart
         * recv
+        * recv_multipart
 
     To ensure that the ``zmq.NOBLOCK`` flag is set and that sending or recieving
     is deferred to the hub if a ``zmq.EAGAIN`` (retry) error is raised.
@@ -48,17 +33,30 @@ class _Socket(_original_Socket):
     """
 
     def __init__(self, context, socket_type):
+        self.__in_send_multipart = False
+        self.__in_recv_multipart = False
         self.__setup_events()
 
-    def close(self):
+    def __del__(self):
+        self.close()
+
+    def close(self, linger=None):
+        super(GreenSocket, self).close(linger)
+        self.__cleanup_events()
+
+    def __cleanup_events(self):
         # close the _state_event event, keeps the number of active file descriptors down
-        if not self._closed and getattr(self, '_state_event', None):
+        if getattr(self, '_state_event', None):
             try:
                 self._state_event.stop()
             except AttributeError, e:
                 # gevent<1.0 compat
                 self._state_event.cancel()
-        super(_Socket, self).close()
+
+        # if the socket has entered a close state resume any waiting greenlets
+        if hasattr(self, '__writable'):
+            self.__writable.set()
+            self.__readable.set()
 
     def __setup_events(self):
         self.__readable = AsyncResult()
@@ -72,13 +70,11 @@ class _Socket(_original_Socket):
             self._state_event = read_event(self.getsockopt(FD), self.__state_changed, persist=True)
 
     def __state_changed(self, event=None, _evtype=None):
+        if self.closed:
+            self.__cleanup_events()
+            return
         try:
-            if self.closed:
-                # if the socket has entered a close state resume any waiting greenlets
-                self.__writable.set()
-                self.__readable.set()
-                return
-            events = self.getsockopt(zmq.EVENTS)
+            events = super(GreenSocket, self).getsockopt(zmq.EVENTS)
         except ZMQError, exc:
             self.__writable.set_exception(exc)
             self.__readable.set_exception(exc)
@@ -90,22 +86,34 @@ class _Socket(_original_Socket):
 
     def _wait_write(self):
         self.__writable = AsyncResult()
-        self.__writable.get()
+        try:
+            self.__writable.get(timeout=1)
+        except gevent.Timeout:
+            self.__writable.set()
 
     def _wait_read(self):
         self.__readable = AsyncResult()
-        self.__readable.get()
+        try:
+            self.__readable.get(timeout=1)
+        except gevent.Timeout:
+            self.__readable.set()
 
     def send(self, data, flags=0, copy=True, track=False):
         # if we're given the NOBLOCK flag act as normal and let the EAGAIN get raised
         if flags & zmq.NOBLOCK:
-            return super(_Socket, self).send(data, flags, copy, track)
+            try:
+                msg = super(GreenSocket, self).send(data, flags, copy, track)
+            finally:
+                if not self.__in_send_multipart:
+                    self.__state_changed()
+            return msg
+
         # ensure the zmq.NOBLOCK flag is part of flags
         flags |= zmq.NOBLOCK
         while True: # Attempt to complete this operation indefinitely, blocking the current greenlet
             try:
                 # attempt the actual call
-                return super(_Socket, self).send(data, flags, copy, track)
+                return super(GreenSocket, self).send(data, flags, copy, track)
             except zmq.ZMQError, e:
                 # if the raised ZMQError is not EAGAIN, reraise
                 if e.errno != zmq.EAGAIN:
@@ -115,12 +123,53 @@ class _Socket(_original_Socket):
 
     def recv(self, flags=0, copy=True, track=False):
         if flags & zmq.NOBLOCK:
-            return super(_Socket, self).recv(flags, copy, track)
+            try:
+                msg = super(GreenSocket, self).recv(flags, copy, track)
+            finally:
+                if not self.__in_recv_multipart:
+                    self.__state_changed()
+            return msg
+
         flags |= zmq.NOBLOCK
         while True:
             try:
-                return super(_Socket, self).recv(flags, copy, track)
+                return super(GreenSocket, self).recv(flags, copy, track)
             except zmq.ZMQError, e:
                 if e.errno != zmq.EAGAIN:
+                    if not self.__in_recv_multipart:
+                        self.__state_changed()
                     raise
+            else:
+                if not self.__in_recv_multipart:
+                    self.__state_changed()
+                return msg
             self._wait_read()
+
+    def send_multipart(self, *args, **kwargs):
+        """wrap send_multipart to prevent state_changed on each partial send"""
+        self.__in_send_multipart = True
+        try:
+            msg = super(GreenSocket, self).send_multipart(*args, **kwargs)
+        finally:
+            self.__in_send_multipart = False
+            self.__state_changed()
+        return msg
+
+    def recv_multipart(self, *args, **kwargs):
+        """wrap recv_multipart to prevent state_changed on each partial recv"""
+        self.__in_recv_multipart = True
+        try:
+            msg = super(GreenSocket, self).recv_multipart(*args, **kwargs)
+        finally:
+            self.__in_recv_multipart = False
+            self.__state_changed()
+        return msg
+
+
+class GreenContext(Context):
+    """Replacement for :class:`zmq.core.context.Context`
+
+    Ensures that the greened Socket above is used in calls to `socket`.
+    """
+    _socket_class = GreenSocket
+
